#include "plugin_module.h"

#include <stdarg.h>

extern int module_sd;
extern FILE *module_fp;
static RecogProcess *cur = NULL;
extern SourceInfo source_info;

#define MAXBUFLEN 4096 			///< Maximum line length of a message sent from a client
static char mbuf[MAXBUFLEN];	///< Work buffer for message output
static char buf[MAXBUFLEN];		///< Work buffer for exec
static char inbuf[MAXBUFLEN];
#ifdef CHARACTER_CONVERSION
static char outbuf[MAXBUFLEN];
#endif

static char *
myfgets(char *buf, int maxlen, FILE *fp)
{
  char *ret;
  int len;
  
  if ((ret = fgets(buf, maxlen, fp)) != NULL) {
    len = strlen(buf);
    if (buf[len-1] == '\n') {
      buf[len-1] = '\0';
      if (len >= 2 && buf[len-2] == '\r') {
        buf[len-2] = '\0';
      }
    }
  }
  return ret;
}

/** 
 * Read grammar (DFA and dictionary) from socket and returns newly allocated
 * grammars.
 * 
 * @param sd [in] socket descpriter
 * @param ret_dfa [out] read DFA
 * @param ret_winfo [out] read dictionary
 * @param hmminfo [in] HMM definition
 * 
 * @return TRUE on success, or FALSE on failure.
 * </EN>
 */
static boolean
read_grammar(FILE *fp, DFA_INFO **ret_dfa, WORD_INFO **ret_winfo, HTK_HMM_INFO *hmminfo, RecogProcess *r)
{
  DFA_INFO *dfa = NULL;
  WORD_INFO *winfo;
  JCONF_LM *lmconf;

  /* load grammar: dfa and dict in turn */
  if (r->lmvar != LM_DFA_WORD) {
    dfa = dfa_info_new();
    if (!rddfa_fp(fp, dfa)) {
      return FALSE;
    }
  }
  winfo = word_info_new();
  if (r->lmvar == LM_DFA_WORD) {
    lmconf = r->lm->config;
    if (!voca_load_wordlist_fp(fp, winfo, hmminfo, lmconf->wordrecog_head_silence_model_name, lmconf->wordrecog_tail_silence_model_name, (lmconf->wordrecog_silence_context_name[0] == '\0') ? NULL : lmconf->wordrecog_silence_context_name)) {
      return FALSE;
    }
  } else {
    if (!voca_load_htkdict_fp(fp, winfo, hmminfo, FALSE)) {
      dfa_info_free(dfa);
      return FALSE;
    }
  }
  *ret_dfa = dfa;
  *ret_winfo = winfo;
  return TRUE;
}


/*Third argument(child) indicates whether SR tag is child or not.*/
static void
send_process_stat(RecogProcess *r, boolean sem_pass,boolean child)
{
#ifdef SEMAPHOR_ON
  if (sem_pass == TRUE){
    seg_begin_wait_semaphor();
  }
#ifdef SEMAPHOR_DEBUG_PASS
  if (sem_pass == TRUE){
    printf("***[SEMAPHOR_DEBUG_PASS]*** No.201[in]\n");
  } else {
    printf("***[SEMAPHOR_DEBUG_PASS]*** No.201[in][no sem]\n");
  }
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
  if(child){
    module_send(module_sd, "  <SR ID=\"%d\" NAME=\"%s\"", r->config->id, r->config->name);
  }else{
    module_send(module_sd, "<SR ID=\"%d\" NAME=\"%s\"", r->config->id, r->config->name);
  }
#ifdef SEMAPHOR_DEBUG_EFFECT
  printf("***[SEMAPHOR_DEBUG_EFFECT]*** No.101<SR>\n");
  fflush(stdout);
  sleep(3);
#endif //SEMAPHOR_DEBUG_EFFECT
  switch(r->lmtype) {
  case LM_PROB: module_send(module_sd, " LMTYPE=\"PROB\""); break;
  case LM_DFA: module_send(module_sd, " LMTYPE=\"DFA\""); break;
  }
  switch(r->lmvar) {
  case LM_NGRAM: module_send(module_sd, " LMVAR=\"NGRAM\""); break;
  case LM_DFA_GRAMMAR: module_send(module_sd, " LMVAR=\"GRAMMAR\""); break;
  case LM_DFA_WORD: module_send(module_sd, " LMVAR=\"WORD\""); break;
  case LM_NGRAM_USER: module_send(module_sd, " LMVAR=\"USER\""); break;
  }
  if (r->live) {
    module_send(module_sd, " LIVE=\"ACTIVE\"");
  } else {
    module_send(module_sd, " LIVE=\"INACTIVE\"");
  }
  if(child){
    module_send(module_sd, "/>\n");
  }else{
    module_send(module_sd, "/>\n.\n");
  }
#ifdef SEMAPHOR_ON
  if (sem_pass == TRUE){
    recog_end_post_semaphor();
  }
#ifdef SEMAPHOR_DEBUG_PASS
  if (sem_pass == TRUE){
    printf("***[SEMAPHOR_DEBUG_PASS]*** No.201[out]\n");
  } else {
    printf("***[SEMAPHOR_DEBUG_PASS]*** No.201[out][no sem]\n");
  }
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
}

static void
send_current_process(RecogProcess *r)
{
#ifdef SEMAPHOR_ON
  seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.202[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
  module_send(module_sd, "<RECOGPROCESS INFO=\"CURRENT\">\n");
  send_process_stat(r, FALSE,TRUE);
#ifdef SEMAPHOR_DEBUG_EFFECT
printf("***[SEMAPHOR_DEBUG_EFFECT]*** No.102<RECOGPROCESS>\n");
fflush(stdout);
sleep(3);
#endif //SEMAPHOR_DEBUG_EFFECT
  module_send(module_sd, "</RECOGPROCESS>\n.\n");
#ifdef SEMAPHOR_ON
  recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.202[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
}

/** 
 * <JA>
 * @brief  ⥸塼륳ޥɤ. 
 *
 * 饤ȤͿ줿ޥɤ. δؿϥ饤Ȥ
 * ޥɤƤ뤿Ӥ˲ǧ˳ǸƤФ. 
 * ơˤĤƤϤǤ˱å뤬
 * ʸˡɲäʤɤϡǤϼΤ߹Ԥºݤѹ
 * ʳʸˡΥХʸˡκƹۤʤɡˤǧι֤˼¹Ԥ. 
 * ʸˡƹ۽ºݤ˹ԤΤ multigram_update() Ǥ. 
 * 
 * @param command [in] ޥʸ
 * </JA>
 * <EN>
 * @brief  Process a module command.
 *
 * This function processes command string received from module client.
 * This will be called whenever a command arrives from a client, interrupting
 * the main recognition process.  The status responses will be performed
 * at this function immediately.  On the whole, grammar modification
 * (add/delete/(de)activation) will not be performed here.  The received
 * data are just stored in this function, and they will be processed later
 * by calling multigram_update() between the recognition process.
 * 
 * @param command [in] command string
 * </EN>
 */
static void
msock_exec_command(char *command, Recog *recog)
{
  DFA_INFO *new_dfa;
  WORD_INFO *new_winfo;
  static char *p, *q;
  int gid;
  int ret;
  RecogProcess *r;

  /* prompt the received command string */
  printf("[[%s]]\n",command);

  if (cur == NULL) {
    cur = recog->process_list;
  }

  if (strmatch(command, "STATUS")) {
#ifdef SEMAPHOR_ON
    seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.203[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
    /* return status */
    if (*(recog->process_active)) {
      module_send(module_sd, "<SYSINFO PROCESS=\"ACTIVE\"/>\n.\n");
    } else {
      module_send(module_sd, "<SYSINFO PROCESS=\"SLEEP\"/>\n.\n");
    }
#ifdef SEMAPHOR_ON
    recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.203[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
  } else if (strmatch(command, "DIE")) {
    /* disconnect */
    close_socket(module_sd);
    module_sd = -1;
#if defined(_WIN32) && !defined(__CYGWIN32__)
    /* this is single process and has not forked, so
       we just disconnect the connection here.  */
#else
    /* this is a forked process, so exit here. */

#endif
  } else if (strmatch(command, "VERSION")) {
#ifdef SEMAPHOR_ON
    seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.204[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
    /* return version */
    module_send(module_sd, "<ENGINEINFO TYPE=\"%s\" VERSION=\"%s\" CONF=\"%s\"/>\n.\n",
                JULIUS_PRODUCTNAME, JULIUS_VERSION, JULIUS_SETUP);
#ifdef SEMAPHOR_ON
    recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.204[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
  } else if (strmatch(command, "PAUSE")) {
    /* pause recognition: will stop when the current input ends */
    j_request_pause(recog);
  } else if (strmatch(command, "TERMINATE")) {
    j_request_terminate(recog);
  } else if (strmatch(command, "RESUME")) {
    j_request_resume(recog);
  } else if (strmatch(command, "INPUTONCHANGE")) {
    /* change grammar switching timing policy */
    if (
      myfgets(buf, MAXBUFLEN, module_fp)
      == NULL) {
      fprintf(stderr, "Error: msock(INPUTONCHANGE): no argument\n");
      return;
    }
    if (strmatch(buf, "TERMINATE")) {
      recog->gram_switch_input_method = SM_TERMINATE;
    } else if (strmatch(buf, "PAUSE")) {
      recog->gram_switch_input_method = SM_PAUSE;
    } else if (strmatch(buf, "WAIT")) {
      recog->gram_switch_input_method = SM_WAIT;
    } else {
      fprintf(stderr, "Error: msock(INPUTONCHANGE): unknown method [%s]\n", buf); exit(-1);
    }
  } else if (strnmatch(command, "GRAMINFO", strlen("GRAMINFO"))) {
    plugin_send_gram_info(cur);
  } else if (strnmatch(command, "CHANGEGRAM", strlen("CHANGEGRAM"))) {
    /* receive grammar (DFA + DICT) from the socket, and swap the whole grammar  */
    /* read grammar name if any */
    p = &(command[strlen("CHANGEGRAM")]);
    while (*p == ' ' && *p != '\r' && *p != '\n' && *p != '\0') p++;
    if (*p != '\r' && *p != '\n' && *p != '\0') {
      q = buf;
      while (*p != ' ' && *p != '\r' && *p != '\n' && *p != '\0') *q++ = *p++;
      *q = '\0';
      p = buf;
    } else {
      p = NULL;
    }
    /* read a new grammar via socket */
    if (read_grammar(module_fp, &new_dfa, &new_winfo, cur->am->hmminfo, cur) == FALSE) {
#ifdef SEMAPHOR_ON
      seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.205[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      module_send(module_sd, "<GRAMMAR STATUS=\"ERROR\" REASON=\"WRONG DATA\"/>\n.\n");
#ifdef SEMAPHOR_ON
      recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.205[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
    } else {
      if (cur->lmtype == LM_DFA) {
        /* delete all existing grammars */
        multigram_delete_all(cur->lm);
        /* register the new grammar to multi-gram tree */
        multigram_add(new_dfa, new_winfo, p, cur->lm);
        /* need to rebuild the global lexicon */
        /* tell engine to update at requested timing */
        schedule_grammar_update(recog);
        /* make sure this process will be activated */
        cur->active = 1;
        /* tell module client  */
#ifdef SEMAPHOR_ON
        seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.206[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
        module_send(module_sd, "<GRAMMAR STATUS=\"RECEIVED\"/>\n.\n");
#ifdef SEMAPHOR_ON
        recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.206[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
        plugin_send_gram_info(cur);
      } else {
#ifdef SEMAPHOR_ON
        seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.207[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
        module_send(module_sd, "<GRAMMAR STATUS=\"ERROR\" REASON=\"NOT A GRAMMAR-BASED LM\"/>\n.\n");
#ifdef SEMAPHOR_ON
        recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.207[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      }
    }
  } else if (strnmatch(command, "ADDGRAM", strlen("ADDGRAM"))) {
    /* receive grammar and add it to the current grammars */
    /* read grammar name if any */
    p = &(command[strlen("ADDGRAM")]);
    while (*p == ' ' && *p != '\r' && *p != '\n' && *p != '\0') p++;
    if (*p != '\r' && *p != '\n' && *p != '\0') {
      q = buf;
      while (*p != ' ' && *p != '\r' && *p != '\n' && *p != '\0') *q++ = *p++;
      *q = '\0';
      p = buf;
    } else {
      p = NULL;
    }
    /* read a new grammar via socket */
    if (read_grammar(module_fp, &new_dfa, &new_winfo, cur->am->hmminfo, cur) == FALSE) {
#ifdef SEMAPHOR_ON
      seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.208[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      module_send(module_sd, "<GRAMMAR STATUS=\"ERROR\" REASON=\"WRONG DATA\"/>\n.\n");
#ifdef SEMAPHOR_ON
      recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.208[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
    } else {
      if (cur->lmtype == LM_DFA) {
        /* add it to multi-gram tree */
        multigram_add(new_dfa, new_winfo, p, cur->lm);
        /* need to rebuild the global lexicon */
        /* make sure this process will be activated */
        cur->active = 1;
        /* tell engine to update at requested timing */
        schedule_grammar_update(recog);
        /* tell module client  */
#ifdef SEMAPHOR_ON
        seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.209[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
        module_send(module_sd, "<GRAMMAR STATUS=\"RECEIVED\"/>\n.\n");
#ifdef SEMAPHOR_ON
        recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.209[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
        plugin_send_gram_info(cur);
      } else {
#ifdef SEMAPHOR_ON
        seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.210[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
        module_send(module_sd, "<GRAMMAR STATUS=\"ERROR\" REASON=\"NOT A GRAMMAR-BASED LM\"/>\n.\n");
#ifdef SEMAPHOR_ON
        recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.210[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      }
    }
  } else if (strmatch(command, "DELGRAM")) {
    /* remove the grammar specified by ID or name */
    /* read a list of grammar IDs to be deleted */
    if (
      myfgets(buf, MAXBUFLEN, module_fp)
      == NULL) {
      fprintf(stderr, "Error: msock(DELGRAM): no argument\n");
      return;
    }
    /* extract IDs and mark them as delete
       (actual deletion will be performed on the next 
    */
    if (cur->lmtype == LM_DFA) {
      for(p=strtok(buf," ");p;p=strtok(NULL," ")) {
        q = p;
        while(*q != '\0' && *q != '\r' && *q != '\n') {
          if (*q < '0' || *q > '9') break;
          q++;
        }
        if (*q == '\0' || *q == '\r' || *q == '\n') { /* numeric */
          gid = atoi(p);
        } else {		/* string */
          gid = multigram_get_id_by_name(cur->lm, p);
          if (gid == -1) continue;
        }
        if (multigram_delete(gid, cur->lm) == FALSE) { /* deletion marking failed */
          fprintf(stderr, "Warning: msock(DELGRAM): gram #%d failed to delete, ignored\n", gid);
          /* tell module */
#ifdef SEMAPHOR_ON
          seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.211[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
          module_send(module_sd, "<GRAMMAR STATUS=\"ERROR\" REASON=\"Gram #%d not found\"/>\n.\n", gid);
#ifdef SEMAPHOR_ON
          recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.211[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
        }
      }
      /* need to rebuild the global lexicon */
      /* tell engine to update at requested timing */
      schedule_grammar_update(recog);
    } else {
#ifdef SEMAPHOR_ON
      seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.212[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      module_send(module_sd, "<GRAMMAR STATUS=\"ERROR\" REASON=\"NOT A GRAMMAR-BASED LM\"/>\n.\n");
#ifdef SEMAPHOR_ON
      recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.212[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
    }
  } else if (strmatch(command, "ACTIVATEGRAM")) {
    /* activate grammar in this engine */
    /* read a list of grammar IDs or names to be activated */
    if (
      myfgets(buf, MAXBUFLEN, module_fp)
      == NULL) {
      fprintf(stderr, "Error: msock(ACTIVATEGRAM): no argument\n");
      return;
    }
    /* mark them as active */
    if (cur->lmtype == LM_DFA) {
      for(p=strtok(buf," ");p;p=strtok(NULL," ")) {
        q = p;
        while(*q != '\0' && *q != '\r' && *q != '\n') {
          if (*q < '0' || *q > '9') break;
          q++;
        }
        if (*q == '\0' || *q == '\r' || *q == '\n') { /* numeric */
          gid = atoi(p);
        } else {		/* string */
          gid = multigram_get_id_by_name(cur->lm, p);
          if (gid == -1) continue;
        }
        ret = multigram_activate(gid, cur->lm);
        if (ret == 1) {
          /* already active */
#ifdef SEMAPHOR_ON
          seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.213[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
          module_send(module_sd, "<WARN MESSAGE=\"Gram #%d already active\"/>\n.\n", gid);
#ifdef SEMAPHOR_ON
          recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.213[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
        } else if (ret == -1) {
          /* not found */
#ifdef SEMAPHOR_ON
          seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.214[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
          module_send(module_sd, "<WARN MESSAGE=\"Gram #%d not found\"/>\n.\n", gid);
#ifdef SEMAPHOR_ON
          recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.214[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
        }	/* else success */
      }
      /* tell engine to update at requested timing */
      schedule_grammar_update(recog);
    } else {
#ifdef SEMAPHOR_ON
      seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.215[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      module_send(module_sd, "<GRAMMAR STATUS=\"ERROR\" REASON=\"NOT A GRAMMAR-BASED LM\"/>\n.\n");
#ifdef SEMAPHOR_ON
      recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.215[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
    }
  } else if (strmatch(command, "DEACTIVATEGRAM")) {
    /* deactivate grammar in this engine */
    /* read a list of grammar IDs or names to be de-activated */
    if (
      myfgets(buf, MAXBUFLEN, module_fp)
      == NULL) {
      fprintf(stderr, "Error: msock(DEACTIVATEGRAM): no argument\n");
      return;
    }
    if (cur->lmtype == LM_DFA) {
      /* mark them as not active */
      for(p=strtok(buf," ");p;p=strtok(NULL," ")) {
        q = p;
        while(*q != '\0' && *q != '\r' && *q != '\n') {
          if (*q < '0' || *q > '9') break;
          q++;
        }
        if (*q == '\0' || *q == '\r' || *q == '\n') { /* numeric */
          gid = atoi(p);
        } else {		/* string */
          gid = multigram_get_id_by_name(cur->lm, p);
          if (gid == -1) continue;
        }
        ret = multigram_deactivate(gid, cur->lm);
        if (ret == 1) {
          /* already inactive */
#ifdef SEMAPHOR_ON
          seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.216[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
          module_send(module_sd, "<WARN MESSAGE=\"Gram #%d already inactive\"/>\n.\n", gid);
#ifdef SEMAPHOR_ON
          recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.216[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
        } else if (ret == -1) {
          /* not found */
#ifdef SEMAPHOR_ON
          seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.217[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
          module_send(module_sd, "<WARN MESSAGE=\"Gram #%d not found\"/>\n.\n", gid);
#ifdef SEMAPHOR_ON
          recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.217[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
        }	/* else success */
      }
      schedule_grammar_update(recog);
    } else {
#ifdef SEMAPHOR_ON
      seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.218[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      module_send(module_sd, "<GRAMMAR STATUS=\"ERROR\" REASON=\"NOT A GRAMMAR-BASED LM\"/>\n.\n");
#ifdef SEMAPHOR_ON
      recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.218[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
    }
  } else if (strmatch(command, "SYNCGRAM")) {
    /* update grammar if necessary */
    if (cur->lmtype == LM_DFA) {
      multigram_update(cur->lm);  /* some modification occured if return TRUE */
      for(r=recog->process_list;r;r=r->next) {
        if (r->lmtype == LM_DFA && r->lm->global_modified) {
          multigram_build(r);
        }
      }
      cur->lm->global_modified = FALSE;
#ifdef SEMAPHOR_ON
      seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.219[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      module_send(module_sd, "<GRAMMAR STATUS=\"READY\"/>\n.\n");
#ifdef SEMAPHOR_ON
      recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.219[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
    } else {
#ifdef SEMAPHOR_ON
      seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.220[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      module_send(module_sd, "<GRAMMAR STATUS=\"ERROR\" REASON=\"NOT A GRAMMAR-BASED LM\"/>\n.\n");
#ifdef SEMAPHOR_ON
      recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.220[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
    }
  } else if (strmatch(command, "CURRENTPROCESS")) {
    JCONF_SEARCH *sconf;
    RecogProcess *r;
    if (
      myfgets(buf, MAXBUFLEN, module_fp)
      == NULL) {
      /* when no argument, just return current process */
      send_current_process(cur);
      return;
    }
    if (buf[0] == '\0') {
      /* when no argument, just return current process */
      send_current_process(cur);
      return;
    }
    sconf = j_get_searchconf_by_name(recog->jconf, buf);
    if (sconf == NULL) {
      fprintf(stderr, "Error: msock(CURRENTPROCESS): no such process \"%s\"\n", buf);
#ifdef SEMAPHOR_ON
      seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.221[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      module_send(module_sd, "<RECOGPROCESS STATUS=\"ERROR\" REASON=\"NO SUCH PROCESS\"/>\n.\n");
#ifdef SEMAPHOR_ON
      recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.221[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      return;
    }
    for(r=recog->process_list;r;r=r->next) {
      if (r->config == sconf) {
        cur = r;
        break;
      }
    }
    if (!r) {
      fprintf(stderr, "Error: msock(CURRENTPROCESS): no process assigned to searchconf \"%s\"??\n", buf);
#ifdef SEMAPHOR_ON
      seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.222[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      module_send(module_sd, "<RECOGPROCESS STATUS=\"ERROR\" REASON=\"NO SUCH PROCESS\"/>\n.\n");
#ifdef SEMAPHOR_ON
      recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.222[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      return;
    }
    send_current_process(cur);
  }

  else if (strmatch(command, "SHIFTPROCESS")) {
    cur = cur->next;
    if (cur == NULL) {
      fprintf(stderr, "SHIFTPROCESS: reached end, rotated to first\n");
      cur = recog->process_list;
    }
    send_process_stat(cur, TRUE,FALSE);
  }

  else if (strmatch(command, "ADDPROCESS")) {
    Jconf *jconf;
    JCONF_LM *lmconf;
    JCONF_AM *amconf;
    JCONF_SEARCH *sconf;
    RecogProcess *r;

    if (
      myfgets(buf, MAXBUFLEN, module_fp)
      == NULL) {
      fprintf(stderr, "Error: msock(ADDPROCESS): no argument\n");
#ifdef SEMAPHOR_ON
      seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.223[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      module_send(module_sd, "<RECOGPROCESS STATUS=\"ERROR\" REASON=\"NO ARGUMENT\"/>\n.\n");
#ifdef SEMAPHOR_ON
      recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.223[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      return;
    }
    /* load specified jconf file and use its last LM conf as new */
    jconf = j_jconf_new();
    j_config_load_file(jconf, buf);
    lmconf = jconf->lmnow;

    /* create a search instance */
    sconf = j_jconf_search_new();
    /* all the parameters are defaults */

    /* create process instance with new LM and SR */
    if (j_process_add_lm(recog, lmconf, sconf, buf) == FALSE) {
      fprintf(stderr, "Error: failed to regist new process \"%s\"\n", buf);
#ifdef SEMAPHOR_ON
      seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.224[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      module_send(module_sd, "<RECOGPROCESS STATUS=\"ERROR\" REASON=\"FAILED TO REGISTER\"/>\n.\n");
#ifdef SEMAPHOR_ON
      recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.224[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      j_jconf_search_free(sconf);
      return;
    }
    printf("added process: SR%02d %s\n", sconf->id, sconf->name);
#ifdef SEMAPHOR_ON
    seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.225[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
    module_send(module_sd, "<RECOGPROCESS INFO=\"ADDED\">\n");
    for(r=recog->process_list;r;r=r->next) {
      if (r->config == sconf) {
        send_process_stat(r, FALSE,TRUE);
      }
    }
#ifdef SEMAPHOR_DEBUG_EFFECT
printf("***[SEMAPHOR_DEBUG_EFFECT]*** No.103<RECOGPROCESS>\n");
fflush(stdout);
sleep(3);
#endif //SEMAPHOR_DEBUG_EFFECT
    module_send(module_sd, "</RECOGPROCESS>\n.\n");
#ifdef SEMAPHOR_ON
    recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.225[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
  }

  else if (strmatch(command, "DELPROCESS")) {
    JCONF_LM *lmconf;
    JCONF_AM *amconf;
    JCONF_SEARCH *sconf;
    RecogProcess *r;

    if (
      myfgets(buf, MAXBUFLEN, module_fp)
      == NULL) {
      fprintf(stderr, "Error: msock(DELPROCESS): no argument\n");
#ifdef SEMAPHOR_ON
      seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.226[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      module_send(module_sd, "<RECOGPROCESS STATUS=\"ERROR\" REASON=\"NO ARGUMENT\"/>\n.\n");
#ifdef SEMAPHOR_ON
      recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.226[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      return;
    }

    lmconf = j_get_lmconf_by_name(recog->jconf, buf);
    if (lmconf == NULL) {
      fprintf(stderr, "Error: msock(DELPROCESS): no lmconf named %s\n", buf);
#ifdef SEMAPHOR_ON
      seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.227[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      module_send(module_sd, "<RECOGPROCESS STATUS=\"ERROR\" REASON=\"NO LM PROCESS OF THE NAME\"/>\n.\n");
#ifdef SEMAPHOR_ON
      recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.227[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      return;
    }
    sconf =  j_get_searchconf_by_name(recog->jconf, buf);
    if (sconf == NULL) {
      fprintf(stderr, "Error: msock(DELPROCESS): no searchconf named %s\n", buf);
#ifdef SEMAPHOR_ON
      seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.228[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      module_send(module_sd, "<RECOGPROCESS STATUS=\"ERROR\" REASON=\"NO RECOGPROCESS OF THE NAME\"/>\n.\n");
#ifdef SEMAPHOR_ON
      recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.228[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      return;
    }
    printf("remove process: SR%02d %s, LM%02d %s\n", sconf->id, sconf->name, lmconf->id, lmconf->name);
#ifdef SEMAPHOR_ON
    seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.229[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
    module_send(module_sd, "<RECOGPROCESS INFO=\"DELETE\">\n");
    for(r=recog->process_list;r;r=r->next) {
      if (r->config == sconf) send_process_stat(r, FALSE,TRUE);
    }
#ifdef SEMAPHOR_DEBUG_EFFECT
printf("***[SEMAPHOR_DEBUG_EFFECT]*** No.104<RECOGPROCESS>\n");
fflush(stdout);
sleep(3);
#endif //SEMAPHOR_DEBUG_EFFECT
    module_send(module_sd, "</RECOGPROCESS>\n.\n");
#ifdef SEMAPHOR_ON
    recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.229[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
    j_process_remove(recog, sconf);
    j_process_lm_remove(recog, lmconf);
    /* change current */
    for(r=recog->process_list;r;r=r->next) {
      if (r == cur) break;
    }
    if (!r) {
      cur = recog->process_list;
      printf("now current moved to SR%02d %s\n", cur->config->id, cur->config->name);
      send_current_process(cur);
    }
    
  }

  else if (strmatch(command, "LISTPROCESS")) {
    RecogProcess *r;
    
#ifdef SEMAPHOR_ON
    seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.230[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
    module_send(module_sd, "<RECOGPROCESS INFO=\"STATUS\">\n");
    for(r=recog->process_list;r;r=r->next) {
      send_process_stat(r, FALSE,TRUE);
    }
#ifdef SEMAPHOR_DEBUG_EFFECT
printf("***[SEMAPHOR_DEBUG_EFFECT]*** No.105<RECOGPROCESS>\n");
fflush(stdout);
sleep(3);
#endif //SEMAPHOR_DEBUG_EFFECT
    module_send(module_sd, "</RECOGPROCESS>\n.\n");
#ifdef SEMAPHOR_ON
    recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.230[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
  }

  else if (strmatch(command, "ACTIVATEPROCESS")) {
    if (
      myfgets(buf, MAXBUFLEN, module_fp)
      == NULL) {
      fprintf(stderr, "Error: msock(ACTIVATEPROCESS): no argument\n");
#ifdef SEMAPHOR_ON
      seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.231[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      module_send(module_sd, "<RECOGPROCESS STATUS=\"ERROR\" REASON=\"NO ARGUMENT\"/>\n.\n");
#ifdef SEMAPHOR_ON
      recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.231[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      return;
    }
    if (j_process_activate(recog, buf) == FALSE) {
#ifdef SEMAPHOR_ON
      seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.232[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      module_send(module_sd, "<RECOGPROCESS STATUS=\"ERROR\" REASON=\"ACTIVATION FAILED\"/>\n.\n");
#ifdef SEMAPHOR_ON
      recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.232[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
    } else {
#ifdef SEMAPHOR_ON
      seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.233[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      module_send(module_sd, "<RECOGPROCESS INFO=\"ACTIVATED\" NAME=\"%s\"/>\n.\n", buf);
#ifdef SEMAPHOR_ON
      recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.233[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
    }
  }
  else if (strmatch(command, "DEACTIVATEPROCESS")) {
    if (
      myfgets(buf, MAXBUFLEN, module_fp)
      == NULL) {
      fprintf(stderr, "Error: msock(DEACTIVATEPROCESS): no argument\n");
#ifdef SEMAPHOR_ON
      seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.234[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      module_send(module_sd, "<RECOGPROCESS STATUS=\"ERROR\" REASON=\"NO ARGUMENT\"/>\n.\n");
#ifdef SEMAPHOR_ON
      recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.234[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      return;
    }
    if (j_process_deactivate(recog, buf) == FALSE) {
#ifdef SEMAPHOR_ON
      seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.235[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      module_send(module_sd, "<RECOGPROCESS STATUS=\"ERROR\" REASON=\"DEACTIVATION FAILED\"/>\n.\n");
    } else {
#ifdef SEMAPHOR_ON
      seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.235[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      module_send(module_sd, "<RECOGPROCESS INFO=\"DEACTIVATED\" NAME=\"%s\"/>\n.\n", buf);
    }
    module_send(module_sd, ".\n");
#ifdef SEMAPHOR_ON
    recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.235[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
  }
  else if (strmatch(command, "ADDWORD")) {
    WORD_INFO *words;
    boolean ret;
    int id;

    /* get gramamr ID to add */
    if (
      myfgets(buf, MAXBUFLEN, module_fp)
      == NULL) {
      fprintf(stderr, "Error: msock(DEACTIVATEPROCESS): no argument\n");
#ifdef SEMAPHOR_ON
      seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.236[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      module_send(module_sd, "<RECOGPROCESS STATUS=\"ERROR\" REASON=\"NO ARGUMENT\"/>\n.\n");
#ifdef SEMAPHOR_ON
      recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.236[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      return;
    }
    id = atoi(buf);

    /* read list of word entries, will stop by "DICEND" */
    words = word_info_new();
    voca_load_start(words, cur->am->hmminfo, FALSE);
    while (
      myfgets(buf, MAXBUFLEN, module_fp)
      != NULL) {
      if (cur->lmvar == LM_DFA_WORD) {
        ret = voca_load_word_line(buf, words, cur->am->hmminfo, 
              cur->lm->config->wordrecog_head_silence_model_name,
              cur->lm->config->wordrecog_tail_silence_model_name,
              (cur->lm->config->wordrecog_silence_context_name[0] == '\0') ? NULL : cur->lm->config->wordrecog_silence_context_name);
      } else {
        ret = voca_load_line(buf, words, cur->am->hmminfo);
      }
      if (ret == FALSE) break;
    }
    ret = voca_load_end(words);
    if (ret == FALSE) {
      fprintf(stderr, "Error: msock(ADDWORD): error in reading word entries\n");
#ifdef SEMAPHOR_ON
      seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.237[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      module_send(module_sd, "<RECOGPROCESS STATUS=\"ERROR\" REASON=\"ERROR IN READING WORD ENTRIES\"/>\n.\n");
#ifdef SEMAPHOR_ON
      recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.237[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      word_info_free(words);
      return;
    }
    if (words->num == 0) {
      fprintf(stderr, "Error: msock(ADDWORD): no word specified\n");
#ifdef SEMAPHOR_ON
      seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.238[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      module_send(module_sd, "<RECOGPROCESS STATUS=\"ERROR\" REASON=\"NO WORD SPECIFIED\"/>\n.\n");
#ifdef SEMAPHOR_ON
      recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.238[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      word_info_free(words);
      return;
    }
    printf("%d words read\n", words->num);
    /* add the words to the grammar */
    if (multigram_add_words_to_grammar_by_id(cur->lm, id, words) == FALSE) {
      fprintf(stderr, "Error: msock(ADDWORD): failed to add words to grammar #%d\n", id);
#ifdef SEMAPHOR_ON
      seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.239[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      module_send(module_sd, "<RECOGPROCESS STATUS=\"ERROR\" REASON=\"FAILED\"/>\n.\n");
#ifdef SEMAPHOR_ON
      recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.239[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
      word_info_free(words);
      return;
    }
    /* book for update */
    schedule_grammar_update(recog);
#ifdef SEMAPHOR_ON
    seg_begin_wait_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.240[in]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
    module_send(module_sd, "%d words added to grammar #%d\n.\n", words->num, id);
    module_send(module_sd, "<RECOGPROCESS INFO=\"ADDEDWORD\" GRAMMARID=\"%d\" NUM=\"%d\"/>\n.\n", id, words->num);
#ifdef SEMAPHOR_ON
    recog_end_post_semaphor();
#ifdef SEMAPHOR_DEBUG_PASS
  printf("***[SEMAPHOR_DEBUG_PASS]*** No.240[out]\n");
#endif //SEMAPHOR_DEBUG_PASS
#endif //SEMAPHOR_ON
    word_info_free(words);
  }
}

/** 
 * <JA>
 * ߥ饤ȥ⥸塼뤫̿᤬Хåեˤ뤫Ĵ١
 * ⤷н. ʤФΤޤ޽λ. 
 * 
 * </JA>
 * <EN>
 * Process one commands from client module.  If no command is in the buffer,
 * it will return without blocking.
 * 
 * </EN>
 */
static void
msock_check_and_process_command(Recog *recog, void *dummy)
{
  fd_set rfds;
  int ret;
  struct timeval tv;

  /* check if some commands are waiting in queue */
  FD_ZERO(&rfds);
  FD_SET(module_sd, &rfds);
  tv.tv_sec = 0;
  tv.tv_usec = 0;      /* 0 msec timeout: return immediately */
  ret = select(module_sd+1, &rfds, NULL, NULL, &tv);
  if (ret < 0) {
    perror("msock_check_and_process_command: cannot poll\n");
  }
  if (ret > 0) {
    /* there is data to read */
    /* process command and change status if necessaty */
    while(1){
      if (myfgets(mbuf, MAXBUFLEN, module_fp) != NULL){
        //printf("mbuf=%s\n",mbuf);
        if (strmatch(mbuf, "SYNCGRAM")){
            //printf("break\n");
            break;
        } else {
          msock_exec_command(mbuf, recog);
        }
      }
    }
  }
}

/** 
 * <JA>
 * 饤ȥ⥸塼뤫̿ɤ߹ǽ. 
 * ̿᤬̵硤ΥޥɤޤԤ. 
 * msock_exec_command()  j_request_resume() ƤФ
 * (*(recog->process_active))  TRUE ˤʤޤǷ֤. 
 * δؿäȤץ resume |!. |
 * 
 * </JA>
 * <EN>
 * Process one commands from client module.  If no command is in the buffer,
 * it will block until next command comes.
 * 
 * </EN>
 */
static void
msock_process_command(Recog *recog, void *dummy)
{

  while(!(*(recog->process_active))) {
    if (
      myfgets(mbuf, MAXBUFLEN, module_fp)
      != NULL) {
      msock_exec_command(mbuf, recog);
    }
  }
}

static void
module_regist_callback(Recog *recog, void *data)
{
  callback_add(recog, CALLBACK_POLL, msock_check_and_process_command, data);
  callback_add(recog, CALLBACK_PAUSE_FUNCTION, msock_process_command, data);
}

void
plugin_module_setup(Recog *recog, void *data)
{
  /* register result output callback functions */
  module_regist_callback(recog, data);
  plugin_setup_output_msock(recog, data);
}
